#ifndef _DB_ACCESS_H
#define _DB_ACCESS_H

/*
 * This file contains the interface to access the database from within
 * abook and filter modules.
 */

/* Prefix for operations within filter modules scope. */
#ifndef F_OP
#define F_OP(x, ...)	x( __VA_ARGS__ )
#endif

enum {
	NAME = 0, /* important */
	EMAIL,
	ADDRESS,
        ADDRESS2,
	CITY,
	STATE,
	ZIP,
	COUNTRY,
	PHONE,
	WORKPHONE,
	FAX,
	MOBILEPHONE,
	NICK,
	URL,
	NOTES,
	ANNIVERSARY,
	ITEM_FIELDS /* this is the last */
};

enum {
	ENUM_ALL,	/* item, field */
	ENUM_SELECTED,	/* item */
	ENUM_STANDARD,	/* field */
	ENUM_EMAILS,	/* field */
};

enum item_ops {
	CREATE,
	RESIZE,
	FREE,
	EMPTY,
	BLANK,
	COPY,
	DUPLICATE,
};

enum field_ops {
	QUERY_KEY,
	QUERY_NAME,
	QUERY_TYPE,
	QUERY_INDEX,
};

typedef char **list_item;

union ret_value {
	int int_v;
	char *str_v;
};

struct db_enumerator {
	int item;
	int mode; /* warning: read only */
};

struct field_enumerator {
	int index;
	int std_id;
	int mode; /* warning: read only */
};


/* Important:
 * 
 * - Database access within abook (and outside database.c and db_access.c)
 *   should be made using the below functions exclusively.
 *
 * - Database access within filter modules should be made using the exported
 *   methods from the 'struct abook_ops' defined in filters/filter.h.
 */

/*
 * Traversal operations.
 */
int real_db_enumerate_items(struct db_enumerator e);
struct db_enumerator init_db_enumerator(int mode);
#define db_next_item(e)		(e.item = F_OP(real_db_enumerate_items, e))
#define db_enumerate_items(e)	while( -1 != db_next_item(e) )

int real_enumerate_fields(struct field_enumerator *e);
struct field_enumerator init_field_enumerator(int mode);
#define next_field(e)		F_OP(real_enumerate_fields, &e)
#define enumerate_fields(e)	while( -1 != next_field(e) )

/*
 * Database commit.
 */
int real_add_item2database(list_item item);
#define add_item2database(item)	F_OP(real_add_item2database, item)

/*
 * Item manipulations.
 */
list_item item_do(enum item_ops op,
		list_item dest, list_item src, list_item *item);
#define item_create()		F_OP(item_do, CREATE, NULL, NULL, NULL)
#define item_resize(item)	F_OP(item_do, RESIZE, NULL, NULL, item)
#define item_free(item)		F_OP(item_do, FREE, NULL, NULL, item)
#define item_empty(item)	F_OP(item_do, EMPTY, item, NULL, NULL)
#define item_blank(item)	F_OP(item_do, BLANK, item, NULL, NULL)
#define item_copy(dest, src)	F_OP(item_do, COPY, dest, src, NULL)
#define item_duplicate(dest, src) F_OP(item_do, DUPLICATE, dest, src, NULL)
int real_item_fput(list_item item, int i, char *val);
#define item_fput(item, i, val)	F_OP(real_item_fput, item, i, val)
char *real_item_fget(list_item item, int i);
#define item_fget(item, i)	F_OP(real_item_fget, item, i)

/*
 * Database read operations.
 */
char *real_db_field_get(int item, int i, int std);
#define db_fget(item, i)		F_OP(real_db_field_get, item, i, 1)
#define db_fget_byid(item, i)		F_OP(real_db_field_get, item, i, 0)
#define db_name_get(item)		F_OP(real_db_field_get, item, NAME, 1)
#define db_email_get(item)		F_OP(real_db_field_get, item, EMAIL, 1)
char *real_get_first_email(int item);
#define get_first_email(item)		F_OP(real_get_first_email, item)
int db_n_items();

/*
 * Database write operations.
 */
int real_db_field_put(int item, int i, int std, char *val);
#define db_fput(item, i, val) 		F_OP(real_db_field_put, item, i, 1, val)
#define db_fput_byid(item, i, val) 	F_OP(real_db_field_put, item, i, 0, val)

/*
 * Field operations.
 */
int real_field_id(int i);
#define field_id(i)	F_OP(real_field_id, i)
union ret_value field_query(enum field_ops action, int i, char *s);
#define field_key(i)	F_OP(field_query, QUERY_KEY, i, NULL).str_v
#define field_name(i)	F_OP(field_query, QUERY_NAME, i, NULL).str_v
#define field_type(i)	F_OP(field_query, QUERY_TYPE, i, NULL).int_v
#define field_index(key) F_OP(field_query, QUERY_INDEX, 0, key).int_v
int real_declare_unknown_field(char *key);
#define declare_unknown_field(key)	F_OP(real_declare_unknown_field, key)


#endif /* _DB_ACCESS_H */
