
/*
 * $Id: filter.c,v 1.50 2005/10/26 19:29:16 jheinonen Exp $
 *
 * by JH <jheinonen@users.sourceforge.net>
 *
 * Copyright (C) Jaakko Heinonen
 * Copyright (C) Cedric Duval
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <ltdl.h>
#ifdef HAVE_CONFIG_H
#	include "config.h"
#endif
#include "abook_curses.h"
#include "abook.h"
#include "db_access.h"
#include "filter.h"
#include "gettext.h"
#include "list.h"
#include "misc.h"
#include "ui.h"
#include "xmalloc.h"
#include <assert.h>


static int get_filter_info(const char *path, lt_ptr data);
static void close_filter(lt_dlhandle handle);


/*
 * Operations available to filters
 */

static void
readonly_ops(struct abook_ops *ops)
{
	ops->real_db_field_get = real_db_field_get;
	ops->real_get_first_email = real_get_first_email;
	ops->real_db_enumerate_items = real_db_enumerate_items;
	ops->real_enumerate_fields = real_enumerate_fields;
	ops->real_field_id = real_field_id;
	ops->field_query = field_query;
	ops->opt_get_value_str = opt_get_value_str;
}

static void
readwrite_ops(struct abook_ops *ops)
{
	ops->real_db_field_put = real_db_field_put;
	ops->real_add_item2database = real_add_item2database;
	ops->item_do = item_do;
	ops->real_item_fget = real_item_fget;
	ops->real_item_fput = real_item_fput;
	ops->real_declare_unknown_field = real_declare_unknown_field;
}


/*
 * External filters
 */

static void
parse_definition_file(FILE *in, char *kw[], abook_list **args)
{
	int i;
	char *line = NULL, *p;

	while(1) {
		free(line);
		if(feof(in) || ((line = p = getaline(in)) == NULL))
			break;
		if(!*line)
			continue;

		SKIPWS(p);
		for(i = 0; kw[i]; i++)
			if(0 == strncmp(kw[i], p, strlen(kw[i]))) {
				p += strlen(kw[i]);
				strtrim(p);
				if(*p) {
					abook_list_append(args, kw[i]);
					abook_list_append(args, p);
				}
				break;
			}
	}
}

static FILE *
try_external(char *path, char *name)
{
	DIR *d;
	struct dirent *e;
	struct stat s;
	FILE *in = NULL;
	char *p;

	p = strdup_printf("%s/%s.external", path, name);
	in = abook_fopen(p, "r");
	free(p);

	if(in || (in = abook_fopen(name, "r")))
		return in;

	if((d = opendir(path)) == NULL)
		return NULL;

	while((e = readdir(d)) != NULL) {
		if(!strcmp(e->d_name, ".") || !strcmp(e->d_name, ".."))
			continue;

		p = strconcat(path, "/", e->d_name, NULL);
		if(!lstat(p, &s) && S_ISDIR(s.st_mode) &&
				((in = try_external(p, name)))) {
			free(p);
			return in;
		}
		free(p);
	}
	closedir(d);

	return NULL;
}

/* Updates props with newly found properties; returns head of those */
static abook_list *
external_filter_properties(char *dir, char *name, abook_list **props)
{
	FILE *in;
	char *keywords[] = { "import_path", "export_path", "description",
		"import_option", "export_option", NULL };
	abook_list *tail, *head;

	for(tail = *props; tail && tail->next; tail = tail->next)
		; /* tail: last before new options */

	if(!(in = try_external(dir, name)))
		return NULL;

	parse_definition_file(in, keywords, props);
	fclose(in);

	if(!tail)
		head = *props;
	else
		head = tail->next;

	if(head) {
		abook_list_append(props, "alias_of");
		abook_list_append(props, name);
		abook_list_append(props, "base_dir");
		abook_list_append(props, dir);
	}
	return head;
}

static void
find_external_filters_in_dir(char *dir, lt_ptr data)
{
	DIR *d;
	struct dirent *e;
	struct stat s;
	char *p, *q;

	if((d = opendir(dir)) == NULL)
		return;

	while((e = readdir(d)) != NULL) {
		if(!strcmp(e->d_name, ".") || !strcmp(e->d_name, ".."))
			continue;

		p = strconcat(dir, "/", e->d_name, NULL);

		if(!lstat(p, &s) && S_ISDIR(s.st_mode))
			find_external_filters_in_dir(p, data);

		if(!(q = strstr(e->d_name, "_filter.external")) ||
				(q == e->d_name)) {
			free(p);
			continue;
		}

		get_filter_info(p, data);

		free(p);
	}
	closedir(d);
}


/*
 * Filter opening
 * 
 * Given a filter name, get the functions and the associated handle
 */
static struct filter_ops *
open_filter(char *name, int mode, abook_list *options, lt_dlhandle *handle)
{
	init_func_t *filter_init;
	struct filter_ops *ops;
	struct abook_ops my_ops;
	abook_list *args = NULL;
	char *rc_dir;

	/*
	 * Searching for a dynamically loadable filter
	 */

	if(lt_dlinit())
		return NULL;

	lt_dladdsearchdir(PATH_FILTERS_IN_BUILD_TREE);
	lt_dladdsearchdir(FILTERS_DIR);

	if((*handle = lt_dlopenext(name)))
		goto found;

	/*
	 * Searching for an external filter
	 */

	rc_dir = strconcat(getenv("HOME"), "/" DIR_IN_HOME, NULL);
	args = external_filter_properties(rc_dir, name, &options);
	free(rc_dir);

	if(!args)
		args = external_filter_properties(FILTERS_DIR, name, &options);

	if(args && (*handle = lt_dlopenext("external_filter")))
		goto found;

	abook_list_free(&options);
	lt_dlexit();
	return NULL; /* Not found */
found:
	/*
	 * Opening filter and fetching exported operations
	 */
	if (!(filter_init = (init_func_t *)lt_dlsym(*handle, "filter_init"))) {
		abook_list_free(&options);
		close_filter(*handle);
		return NULL;
	}

	memset(&my_ops, 0, sizeof(struct abook_ops));
	switch(mode) {
		case F_IMPORT: readwrite_ops(&my_ops);
			       /* fall through */
		case F_EXPORT: readonly_ops(&my_ops);
	}

	ops = (*filter_init)(&my_ops, options);
	abook_list_free(&options);

	if(!ops)
		close_filter(*handle);

	return ops;
}

static void
close_filter(lt_dlhandle handle)
{
	close_func_t *close_func = (close_func_t *)lt_dlsym(handle,
			"filter_close");

	if(close_func)
		(*close_func)();

	lt_dlclose(handle);
	lt_dlexit();
}


/*
 * List of available filters
 */

static int
get_filter_info(const char *path, lt_ptr data)
{
	lt_dlhandle handle;
	struct filter_ops *ops;
	struct filters_info *infos = (struct filters_info *)data;
	char *str, *p, *q, *name, *desc;

	if(!(p = strstr(path, "_filter")))
		return 0;

	if(!(ops = open_filter((char *)path, F_NONE, NULL, &handle)))
		return 0;

	if(!(desc = ops->description()) || !*desc)
		return 0; /* filter does not want to be listed */

	q = strrchr(path, '/');
	q = q ? q + 1 : (char *)path;
	name = xstrndup(q, p - q);

	str = strdup_printf("%s\t%s%s",
			name,
			strstr(path, ".external") ? _("[external] ") : "",
			gettext(desc));
	free(name);

	if(ops->import)
		abook_list_insert_sorted(&infos->input, str);

	if(ops->export)
		abook_list_insert_sorted(&infos->output, str);

	free(str);
	close_filter(handle);

	return 0;
}

static void
filters_info(struct filters_info *parms)
{
	struct stat s;
	char *rc_dir;

	/*
	 * Dynamically loadable filters
	 */
	if(lt_dlinit())
		return;

	if(!lstat(PATH_FILTERS_IN_BUILD_TREE, &s) && S_ISDIR(s.st_mode))
		lt_dlforeachfile(PATH_FILTERS_IN_BUILD_TREE,
				get_filter_info, (lt_ptr)parms);
	else
		lt_dlforeachfile(FILTERS_DIR, get_filter_info, (lt_ptr)parms);

	lt_dlexit();

	/*
	 * External filters
	 */
	rc_dir = strconcat(getenv("HOME"), "/" DIR_IN_HOME, NULL);
	find_external_filters_in_dir(rc_dir, (lt_ptr)parms);
	free(rc_dir);
	find_external_filters_in_dir(FILTERS_DIR, (lt_ptr)parms);
}

void
print_filters()
{
	abook_list *cur;
	struct filters_info infos = { NULL, NULL };

	filters_info(&infos);

	puts(_("input:"));
	for(cur = infos.input; cur; cur = cur->next)
		printf("\t%s\n", cur->data);

	putchar('\n');

	puts(_("output:"));
	for(cur = infos.output; cur; cur = cur->next)
		printf("\t%s\n", cur->data);

	abook_list_free(&infos.input);
	abook_list_free(&infos.output);
}

static char *
filters_screen(enum filter_actions action)
{
	int i, filter;
	char *str = NULL, *p;
	struct filters_info infos = { NULL, NULL };
	abook_list *cur, *list;

	clear();
	refresh_statusline();

	filters_info(&infos);

	switch(action) {
		case F_IMPORT:
			headerline(_("import database"));
			list = infos.input;
			abook_list_free(&infos.output);
			break;
		case F_EXPORT:
			headerline(_("export database"));
			list = infos.output;
			abook_list_free(&infos.input);
			break;
		default: assert(0);
	}

	mvaddstr(3, 1, _("please select a filter"));

	for(cur = list, i = 0; cur; cur = cur->next, i++)
		mvprintw(5 + i, 6, "%c -\t%s\n", 'a' + i, cur->data);

	mvprintw(6 + i, 6, _("x -\tcancel"));

	do {
		if((filter = tolower(getch()) - 'a') == ('x' - 'a')) {
			refresh_screen();
			goto out;
		}
	} while((filter >= abook_list_count(list)) || (filter < 0));

	mvaddstr(5 + filter, 2, "->");

	str = p = abook_list_get(list, filter);
	if(str) {
		SKIPNWS(p);
		str = xstrndup(str, p - str);
	}

out:
	abook_list_free(&list);
	return str;
}


/*
 * import
 */

static int		i_read_file(char *filename, int (*func) (FILE *in));

static import_func_t *
import_func(char *filtname, abook_list *options, lt_dlhandle *handle)
{
	struct filter_ops *ops;
	char *name = strdup_printf("%s_filter", filtname);

	ops = open_filter(name, F_IMPORT, options, handle);
	free(name);
	if(!ops || !ops->import)
		return NULL;

	return ops->import;
}
	
void
import_database()
{
	char *filename, *filtname;
	int tmp = db_n_items();
	lt_dlhandle handle;
	import_func_t *func;

	if(!(filtname = filters_screen(F_IMPORT)))
		goto out_import1;

	if(!(filename = ask_filename(_("Filename: "))))
		goto out_import1;

	if(!(func = import_func(filtname, NULL, &handle)))
		goto out_import2;

	if(i_read_file(filename, func))
		statusline_msg(_("Error occured while opening the file"));
	else if(tmp == db_n_items())
		statusline_msg(_("File does not seem to be a valid addressbook"));

	free(filename);
out_import2:
	close_filter(handle);
out_import1:
	refresh_screen();
}

static int
i_read_file(char *filename, int (*func) (FILE *in))
{
	FILE *in;
	int ret = 0;

	if( (in = abook_fopen( filename, "r" )) == NULL )
		return 1;

	ret = (*func) (in);

	fclose(in);

	return ret;
}

int
import_file(char filtname[FILTNAME_LEN], char *filename, abook_list *options)
{
	int ret = 0, tmp = db_n_items();
	lt_dlhandle handle;
	import_func_t *func;
	
	if((func = import_func(filtname, options, &handle)) == NULL)
		return 1;
	
	if(!strcmp(filename, "-")) {
		struct stat s;
		if((fstat(fileno(stdin), &s)) == -1 || S_ISDIR(s.st_mode))
			ret = 1;
		else
			ret = (*func) (stdin);
	} else
		ret = i_read_file(filename, func);

	if(tmp == db_n_items())
		ret = 1;

	close_filter(handle);

	return ret;
}


/*
 * export
 */

static int e_write_file(char *filename,
		int (*func) (FILE *in, struct db_enumerator e), int mode);

static export_func_t *
export_func(char *filtname, abook_list *options, lt_dlhandle *handle)
{
	struct filter_ops *ops;
	char *name = strdup_printf("%s_filter", filtname);

	ops = open_filter(name, F_EXPORT, options, handle);
	free(name);
	if(!ops || !ops->export)
		return NULL;

	return  ops->export;
}

void
export_database()
{
	int enum_mode = ENUM_ALL;
	char *filename, *filtname;
	lt_dlhandle handle;
	export_func_t *func;

	if(!(filtname = filters_screen(F_EXPORT)))
		goto out_export1;

	if(selected_items()) {
		switch(statusline_askchoice(
			_("Export <a>ll, export <s>elected, or <c>ancel?"),
			S_("keybindings:all/selected/cancel|asc"), 3)) {
			case 1:	break;
			case 2:	enum_mode = ENUM_SELECTED; break;
			case 0: /* fall through */ case 3: goto out_export1;
		}
		clear_statusline();
	}

	if(!(filename = ask_filename(_("Filename: "))))
		goto out_export1;

	if(!(func = export_func(filtname, NULL, &handle)))
		goto out_export2;

	if(e_write_file(filename, func, enum_mode))
		statusline_msg(_("Error occured while exporting"));

	close_filter(handle);
out_export2:
	free(filename);
out_export1:
	refresh_screen();
}

static int
e_write_file(char *filename, int (*func) (FILE *in, struct db_enumerator e),
		int mode)
{
	FILE *out;
	int ret = 0;
	struct db_enumerator enumerator = init_db_enumerator(mode);

	if((out = fopen(filename, "a")) == NULL)
		return 1;

	if(ftell(out))
		return 1;

	ret = (*func) (out, enumerator);

	fclose(out);

	return ret;
}

int
fexport(char filtname[FILTNAME_LEN], FILE *fp, int enum_mode)
{
	int ret;
	struct db_enumerator e = init_db_enumerator(enum_mode);
	lt_dlhandle handle;
	export_func_t *func;

	if((func = export_func(filtname, NULL, &handle)) == NULL)
		return -1;

	ret = (*func) (fp, e);

	close_filter(handle);

	return ret;
}

int
export_file(char filtname[FILTNAME_LEN], char *filename, abook_list *options)
{
	const int mode = ENUM_ALL;
	int ret = 0;
	struct db_enumerator e = init_db_enumerator(mode);
	lt_dlhandle handle;
	export_func_t *func;

	if((func = export_func(filtname, options, &handle)) == NULL)
		return -1;

	if(!strcmp(filename, "-"))
		ret = (*func) (stdout, e);
	else
		ret =  e_write_file(filename, func, mode);

	close_filter(handle);

	return ret;
}


