
/*
 * $Id$
 *
 * by Cedric Duval <cedricduval@free.fr>
 *
 * Copyright (C) Cedric Duval
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#ifdef HAVE_CONFIG_H
#	include "../config.h"
#endif
#include "../abook.h"
#include "../db_access.h"
#include "../misc.h"
#include "../xmalloc.h"
#include "assert.h"

#include "filter_interface.h"
#include "filter.h"

#include "external.h"

static char *i_path = NULL, *e_path = NULL, *base_dir = NULL;
static char *filter_desc = NULL, *alias_of = NULL;
static abook_list *i_options = NULL, *e_options = NULL;

static FILE *in_cmd, *out_data, *in_data;
static list_item item;
static int send_errors;
static struct db_enumerator db_enum;
static struct field_enumerator f_enum;


static void cmd_put_field(abook_buffer *b);
static void cmd_get_field(abook_buffer *b);
static void cmd_enumerate(abook_buffer *b);
static void cmd_emails(abook_buffer *b);
static void cmd_options(abook_buffer *b);
static void cmd_errors(abook_buffer *b);


/*
 * Important:
 *  - keep this array in sync with the corresponding enum in external.h
 *  - any addition or change made to the commands must be reflected
 *    in the doc/README.external_filters documentation
 */
static struct {
	char *cmd;
	int access;
	void (*func) (abook_buffer *);
} cmds[] = {
	{ "errors",	ACCESS_BOTH,	cmd_errors }, /* CMD_ERRORS */
	{ "stop",	ACCESS_BOTH,	NULL }, /* CMD_STOP */
	{ "puts",	ACCESS_BOTH,	NULL }, /* CMD_PUTS */
	{ "options",	ACCESS_BOTH,	cmd_options }, /* CMD_OPTIONS */
	{ "abook_option", ACCESS_BOTH,	NULL }, /* CMD_ABOOK_OPTION */

	{ "next_line",	ACCESS_IN,	NULL }, /* CMD_NEXT_LINE */
	{ "field",	ACCESS_IN,	cmd_put_field }, /* CMD_PUT_FIELD */
	{ "commit_item", ACCESS_IN,	NULL }, /* CMD_COMMIT_ITEM */

	{ "next_item",	ACCESS_OUT,	NULL }, /* CMD_NEXT_ITEM */
	{ "field",	ACCESS_OUT,	cmd_get_field }, /* CMD_GET_FIELD */
	{ "field_name",	ACCESS_OUT,	NULL }, /* CMD_FIELD_NAME */
	{ "enumerate",	ACCESS_OUT,	cmd_enumerate }, /* CMD_ENUMERATE */
	{ "emails",	ACCESS_OUT,	cmd_emails }, /* CMD_EMAILS */

	{ NULL }
};

/*
 * Important:
 *   - keep this array in sync with the corresponding enum in external.h
 *   - do not insert or remove, only append OK or error codes, in order to
 *     preserve compatibility of code numbers
 */
static char *errors[] = {
	"OK", /* ERR_OK */
	"OK - new field declared", /* ERR_OK_DECLARED_NEW_FIELD */
	/* ERR_OK_SEPARATOR - error numbers starts from 1 with next entry */
	"invalid value", /* ERR_INVALID_VALUE */
	"invalid option", /* ERR_INVALID_OPTION */
	"invalid command", /* ERR_INVALID_CMD */
	"command not authorized in this mode", /* ERR_CMD_NOT_AUTHORIZED */
	"malformed request", /* ERR_MALFORMED_REQUEST */
	"field undeclared", /* ERR_UNDECLARED_FIELD */
	"field name is mandatory", /* ERR_NO_NAME_FIELD */
};


static int send_error_str(int err, char *s)
{
	int i;

	i = (err < ERR_OK_SEPARATOR) ? err : err - 1;
	assert(i >= 0 && i < sizeof(errors) / sizeof(char *));
	err = (err < ERR_OK_SEPARATOR) ? 0 : err - ERR_OK_SEPARATOR;

	if(send_errors == SEND_ERR_NONE)
		return -err;


	switch(send_errors) {
		case SEND_ERR_STDERR:
			if(err != 0)
				fprintf(stderr, send_error_string(err, s, i));
			break;
		case SEND_ERR_FILTER:
			fprintf(out_data, send_error_string(err, s, i));
			fflush(out_data);
			break;
	}

	return -err;
}
#define send_error(err) send_error_str(err, NULL)

/*
 * Launching the external filter
 */
static int
external_run(char *path, FILE *out)
{
	int fd_command[2], fd_data[2];
	pid_t child;

	if((pipe(fd_command) == -1) || (pipe(fd_data) == -1) ||
			((child = fork()) == -1))
		return -1;

	if(child == -1) {
		fprintf(stderr, "fork failed\n");
		return -1;
	} if(child == 0) {
		int cmd_fd;
		char *s;

		close(fd_data[1]);
		close(fd_command[0]);

		close(STDERR_FILENO);
		cmd_fd = dup2(fd_command[1], STDERR_FILENO);
		close(fd_command[1]);

		if(out && out != stdout) {
			close(STDOUT_FILENO);
			dup2(fileno(out), STDOUT_FILENO);
			close(fileno(out));
		}

		close(STDIN_FILENO);
		dup2(fd_data[0], STDIN_FILENO);
		close(fd_data[0]);

		execlp(path, path, NULL);

		s = strdup_printf("stop '%s': execution failed.\n", path);
		write(cmd_fd, s, strlen(s));
		free(s);
		fsync(cmd_fd);
		exit(1);
	}

	in_cmd = fdopen(fd_command[0], "r");
	out_data = fdopen(fd_data[1], "w");
	close(fd_data[0]);
	close(fd_command[1]);

	return child;
}


/*
 * Handling of various commands
 */

static void
cmd_put_field(abook_buffer *b)
{
	int field, ret;
	static int preserve = -1;
	
	if(preserve == -1) {
		char *s = opt_get_value_string("preserve_fields");
		preserve = !strcasecmp(s, "all") ? 1 : 0;
		free(s);
	}

	if(get_token(b, TOKEN_EQUAL) || !b->data) {
		send_error(ERR_MALFORMED_REQUEST);
		return;
	}

	if((field = field_index(b->data)) != -1) {
		if(item[field])
			free(item[field]);
		if(*b->ptr)
			item[field] = xstrdup(b->ptr);
		ret = ERR_OK;
	} else if(preserve) {
		if(!*b->ptr) {
			send_error(ERR_MALFORMED_REQUEST);
			return;
		}
		field = declare_unknown_field(b->data);
		assert(field != -1);
		item_resize(&item);
		item[field] = xstrdup(b->ptr);
		ret = ERR_OK_DECLARED_NEW_FIELD;
	} else
		ret = ERR_UNDECLARED_FIELD;

	send_error(ret);
}

static void
cmd_get_field(abook_buffer *b)
{
	int i;
	char *str;

        for(; !get_token(b, TOKEN_COMMA) && b->data; )
		if((i = field_index(b->data)) != -1)
			if((str = db_fget_byid(db_enum.item, i)) && *str) {
				send_data(str);
				return;
			}
	send_data("");
}

static void
cmd_enumerate(abook_buffer *b)
{
	int all = 0, send_mode_equal = 0, sent = 0;
	char *str;

	while(!get_token(b, TOKEN_COMMA) && b->data) {
		if(!strcasecmp(b->data, "all"))
			all = 1;
		else if(!strcasecmp(b->data, "equal"))
			send_mode_equal = 1;
	}

	while(sent == 0) {
		if(next_field(f_enum) == -1) {
			send_data(""); /*empty string indicates end of fields*/
			return;
		}

		str = db_fget_byid(db_enum.item, f_enum.index);
		if((str && *str) || all) {
			if(send_mode_equal)
				send_data_fmt("%s=%s\n",
						field_key(f_enum.index),
						safe_str(str));
			else {
				send_data(field_key(f_enum.index));
				send_data(safe_str(str));
			}
			sent = 1;
		}
	}
}

static void
cmd_errors(abook_buffer *b)
{
	if(get_token(b, 0) || !b->data) {
		send_error(ERR_MALFORMED_REQUEST);
		return;
	}

	if(!strcasecmp(b->data, "none"))
		send_errors = SEND_ERR_NONE;
	else if(!strcasecmp(b->data, "stderr"))
		send_errors = SEND_ERR_STDERR;
	else if(!strcasecmp(b->data, "filter"))
		send_errors = SEND_ERR_FILTER;
	else {
		send_error_str(ERR_INVALID_VALUE, "errors");
		return;
	}

	send_error(ERR_OK);
}

static void
cmd_emails(abook_buffer *b)
{
	struct field_enumerator fe = init_field_enumerator(ENUM_EMAILS);
	abook_list *emails, *cur;

	enumerate_fields(fe) {
		emails = csv_to_abook_list(db_fget_byid(db_enum.item,fe.index));
		if(emails) {
			for(cur = emails; cur; cur = cur->next)
				send_data(cur->data);
			abook_list_free(&emails);
		}
	}
	send_data(""); /* empty string indicates end of list */
}

static void
cmd_options(abook_buffer *b)
{
	abook_list *cur = in_data ? i_options : e_options;

	for(; cur; cur = cur->next)
		send_data(cur->data);
	send_data(""); /* void string indicates end of list */
}


/*
 * Handling of the external process requests
 */

static int
get_command(abook_buffer *b)
{
	int i, ret = 0;

	if(get_token(b, 0) || !b->data)
		return -1;

	for(i = 0; cmds[i].cmd; i ++)
		if(!strcasecmp(b->data, cmds[i].cmd)) {
			if((in_data && (cmds[i].access == ACCESS_OUT)) ||
				(!in_data && (cmds[i].access == ACCESS_IN))) {
				ret = -2; /* command not authorized */
				continue;
			}

			if(cmds[i].func) {
				cmds[i].func(b);
				return -1; /* processed */
			} else
				return i; /* still to be processed */
		}
	return ret ? ret : -3;
}

static int
process_command(abook_buffer *b)
{
	char *str;
	int cmd, i;

	switch(cmd = get_command(b)) {
		case CMD_NEXT_LINE:
			if(!feof(in_data) && (str = getaline(in_data)) != NULL){
				send_data(str);
			} else
				send_data("_ABOOK_DATA_EOF_");
			break;
		case CMD_COMMIT_ITEM:
			if(!item[field_id(NAME)]) {
				item_empty(item);
				send_error(ERR_NO_NAME_FIELD);
			} else {
				add_item2database(item);
				item_blank(item);
				send_error(ERR_OK);
			}
			break;
		case CMD_NEXT_ITEM:
			if(db_next_item(db_enum) == -1) {
				send_data("");
				return 1;
			}
			f_enum = init_field_enumerator(ENUM_ALL);
			send_data(db_name_get(db_enum.item));
			break;
		case CMD_FIELD_NAME:
			i = field_index(b->ptr);
			if(i != -1)
				send_data(safe_str(field_name(i)));
			else
				send_data("");
			break;
		case CMD_ABOOK_OPTION:
			if((str = opt_get_value_string(b->ptr))) {
				send_data(str);
				free(str);
			} else
				send_error_str(ERR_INVALID_OPTION, b->ptr);
			break;
		case CMD_STOP:
		case CMD_PUTS: /* fall through */
			if(*b->ptr)
				fprintf(stderr, "%s\n", b->ptr);
			if(cmd == CMD_STOP)
				return 1;
			break;
		case -2:
			send_error_str(ERR_CMD_NOT_AUTHORIZED, b->data);
			break;
		case -3:
			send_error_str(ERR_INVALID_CMD, b->data);
			break;
		default:
			assert(cmd <= 0); /*otherwise, unhandled known command*/
	}
	return 0;
}

static void
process_external_commands()
{
	char *line = NULL;
	abook_buffer b;

	while(1) {
		free(line);
		if((line = b.ptr = getaline(in_cmd)) == NULL)
			goto out;

		if(feof(in_cmd))
			break;

		strtrim(b.ptr);
		if(!*b.ptr)
			continue;

		if(process_command(&b))
			break;
	}

	free(line);
out:
	fclose(in_cmd);
	fclose(out_data);
}


/*
 * Export filter
 */
static int
external_export_database(FILE *out, struct db_enumerator e)
{
	pid_t child;
	if(!e_path)
		return 1;

	child = external_run(e_path, out);

	if(child < 0)
		return 1;

	db_enum = e;
	if(db_next_item(db_enum) == -1) /* advancing to first item */
		return 1;
	f_enum = init_field_enumerator(ENUM_ALL); /* first field */

	process_external_commands();

	waitpid(child, NULL, 0);

	return 0;
}


/*
 * Import filter
 */
static int
external_parse_file(FILE *in)
{
	pid_t child;

	if(!i_path)
		return 1;

	child = external_run(i_path, NULL);

	if(child < 0)
		return 1;

	in_data = in;
	item = item_create();

	process_external_commands();

	waitpid(child, NULL, 0);

	item_empty(item);
	item_free(&item);
	
	return 0;
}


/*
 * Some (de)initializations
 */
static void
external_filter_init(abook_list *args)
{
	abook_list *a;

	for(a = args; a; a = a->next->next) {
		if(!strncmp(a->data, "description", 11))
			filter_desc = xstrdup(a->next->data);
		else if(!strncmp(a->data, "import_path", 11))
			i_path = a->next->data;
		else if(!strncmp(a->data, "export_path", 11))
			e_path = a->next->data;
		else if(!strncmp(a->data, "alias_of", 8))
			alias_of = xstrdup(a->next->data);
		else if(!strncmp(a->data, "base_dir", 8))
			base_dir = xstrdup(a->next->data);
		else if(!strncmp(a->data, "import_option", 13))
			abook_list_append(&i_options, a->next->data);
		else if(!strncmp(a->data, "export_option", 13))
			abook_list_append(&e_options, a->next->data);
	}

	if(i_path) {
		if((*i_path != '/') && base_dir)
			i_path = strconcat(base_dir, "/", i_path, NULL);
		else
			i_path = xstrdup(i_path);
	}

	if(e_path) {
		if((*e_path != '/') && base_dir)
			e_path = strconcat(base_dir, "/", e_path, NULL);
		else
			e_path = xstrdup(e_path);
	}

	send_errors = SEND_ERR_STDERR;
}

void
external_close_filter()
{
	xfree(filter_desc);
	xfree(i_path);
	xfree(e_path);
	xfree(alias_of);
	xfree(base_dir);
	abook_list_free(&i_options);
	abook_list_free(&e_options);
}

static char *
external_filter_description()
{
	return safe_str(filter_desc);
}


/*
 * Using filter helper
 */
#define FILTER_NAME external
#define FILTER_EXPORT_FUNCTION external_export_database
#define FILTER_IMPORT_FUNCTION external_parse_file
#define FILTER_CLOSE_FUNCTION external_close_filter
#define FILTER_INIT_HOOK external_filter_init
#define FILTER_DESC_HOOK external_filter_description
#include "filter_helper.h"
