
/*
 * $Id$
 *
 * (C) by abook developers
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_CONFIG_H
#	include "../config.h"
#endif
#include "../abook.h"
#include "../db_access.h"
#include "../misc.h"
#include "../xmalloc.h"
#include <assert.h>

#include "ldif.h"

#include "filter_interface.h"
#include "filter.h"


/*
 * ldif import
 */

#define	LDIF_ITEM_FIELDS	16

typedef char *ldif_item[LDIF_ITEM_FIELDS];

static ldif_item ldif_field_names = {
	"cn",
	"mail",
	"streetaddress",
	"streetaddress2",
        "locality",
	"st",
	"postalcode",
	"countryname",
	"homephone",
	"description",
	"homeurl",
	"facsimiletelephonenumber",
	"cellphone",
	"xmozillaanyphone",
	"xmozillanickname",
	"objectclass", /* this must be the last entry */
};

static int ldif_conv_table[LDIF_ITEM_FIELDS] = {
	NAME,		/* "cn" */
	EMAIL,		/* "mail" */
	ADDRESS,	/* "streetaddress" */
	ADDRESS2,	/* "streetaddress2" */
        CITY,		/* "locality" */
	STATE,		/* "st" */
	ZIP,		/* "postalcode" */
	COUNTRY,	/* "countryname" */
	PHONE,		/* "homephone" */
	NOTES,		/* "description" */
	URL,		/* "homeurl" */
	FAX,		/* "facsimiletelephonenumber" */
	MOBILEPHONE,	/* "cellphone" */
	WORKPHONE,	/* "xmozillaanyphone" */
	NICK,		/* "xmozillanickname" */
	-1,             /* "objectclass" */ /* this must be the last entry */
};

static char *
ldif_read_line(FILE *in)
{
	char *buf = NULL;
	char *ptr, *tmp;
	long pos;
	int i;

	for(i = 1; ; i++) {
		char *line;

		pos = ftell(in);
		line = getaline(in);

		if(feof(in) || !line)
			break;

		if(i == 1) {
			buf = line;
			continue;
		}

		if(*line != ' ') {
			fseek(in, pos, SEEK_SET); /* fixme ! */
			free(line);
			break;
		}

		ptr = line;
		while( *ptr == ' ')
			ptr++;

		tmp = buf;
		buf = strconcat(buf, ptr, NULL);
		free(tmp);
		free(line);
	}

	if(buf && *buf == '#' ) {
		free(buf);
		return NULL;
	}

	return buf;
}

static void
ldif_add_item(ldif_item li)
{
	list_item abook_item;
	int i, j;

	abook_item = item_create();

	if(!li[LDIF_ITEM_FIELDS - 1])
		goto bail_out;

	for(i = 0; i < LDIF_ITEM_FIELDS; i++) {
		if((j = ldif_conv_table[i]) < 0 || !li[i] || !*li[i])
			continue;

		item_fput(abook_item, j, xstrdup(li[i]));
	}

	add_item2database(abook_item);

bail_out:
	for(i = 0; i < LDIF_ITEM_FIELDS; i++)
		xfree(li[i]);
	item_free(&abook_item);

}

static void
ldif_convert(ldif_item item, char *type, char *value)
{
	int i;

	if(!strcmp(type, "dn")) {
		ldif_add_item(item);
		return;
	}

	for(i = 0; i < LDIF_ITEM_FIELDS; i++) {
		if(!safe_strcmp(ldif_field_names[i], type) && *value) {
			if(i == LDIF_ITEM_FIELDS - 1) /* this is a dirty hack */
				if(safe_strcmp("person", value))
					break;
			if(item[i])
				xfree(item[i]);
			item[i] = xstrdup(value);
		}
	}
}

static void
ldif_fix_string(char *str)
{
	int i, j;

	for(i = 0, j = 0; j < (int)strlen(str); i++, j++)
		str[i] = ( str[j] == (char)0xc3 ?
				(char) str[++j] + (char) 0x40 :
				str[j] );

	str[i] = 0;
}

static int
ldif_parse_file(FILE *handle)
{
	char *line = NULL;
	char *type, *value;
	int vlen;
	ldif_item item;

	memset(item, 0, sizeof(item));

	do {
		if( !(line = ldif_read_line(handle)) )
			continue;

		if(-1 == (str_parse_line(line, &type, &value, &vlen))) {
			xfree(line);
			continue; /* just skip the errors */
		}

		ldif_fix_string(value);

		ldif_convert(item, type, value);

		xfree(line);
	} while ( !feof(handle) );

	ldif_convert(item, "dn", "");

	return 0;
}

/*
 * end of ldif import
 */



/*
 * ldif export filter
 */

static void
ldif_fput_type_and_value(FILE *out,char *type, char *value )
{
	char *tmp;

	tmp = ldif_type_and_value(type, value, strlen(value));

	fputs(tmp, out);

	free(tmp);
}

static int
ldif_export_database(FILE *out, struct db_enumerator e)
{
	char *email;

	fprintf(out, "version: 1\n");

	db_enumerate_items(e) {
		char *tmp;
		int j;
		email = get_first_email(e.item);

		tmp = strdup_printf("cn=%s,mail=%s",db_name_get(e.item),email);

		ldif_fput_type_and_value(out, "dn", tmp);
		free(tmp);

		for(j = 0; j < LDIF_ITEM_FIELDS; j++) {
			if(ldif_conv_table[j] >= 0) {
				if(ldif_conv_table[j] == EMAIL)
					ldif_fput_type_and_value(out,
						ldif_field_names[j], email);
				else if(db_fget(e.item,ldif_conv_table[j]))
					ldif_fput_type_and_value(out,
						ldif_field_names[j],
						db_fget(e.item,
							ldif_conv_table[j]));
			}
		}
		xfree(email);

		fprintf(out, "objectclass: top\n"
				"objectclass: person\n\n");
	}

	return 0;
}

/*
 * end of ldif export
 */



/*
 * Using filter helper
 */
#define FILTER_NAME ldif
#define FILTER_DESCRIPTION N_("ldif / Netscape addressbook (.4ld)")
#define FILTER_IMPORT_FUNCTION ldif_parse_file
#define FILTER_EXPORT_FUNCTION ldif_export_database
#include "filter_helper.h"
